package flink.zanxiangnet.ad.monitoring;

import com.zanxiangnet.module.util.DateUtil;
import flink.zanxiangnet.ad.monitoring.kafka.KafkaComponent;
import flink.zanxiangnet.ad.monitoring.pojo.dto.AdStatOfDayODSDTO;
import flink.zanxiangnet.ad.monitoring.pojo.entity.AdDataOfDayODS;
import flink.zanxiangnet.ad.monitoring.pojo.entity.AdStatOfDayDWD;
import flink.zanxiangnet.ad.monitoring.pojo.properties.ApplicationProperties;
import flink.zanxiangnet.ad.monitoring.process.AdDayDWDRollMonthProcess;
import flink.zanxiangnet.ad.monitoring.process.AdDayDWDRollYearProcess;
import flink.zanxiangnet.ad.monitoring.sink.AdDayDWDToDBBatchSink;
import flink.zanxiangnet.ad.monitoring.sink.OssBatchStreamSink;
import flink.zanxiangnet.ad.monitoring.stream.BatchStream;
import flink.zanxiangnet.ad.monitoring.stream.KeyedBatchStream;
import lombok.extern.slf4j.Slf4j;
import org.apache.commons.lang3.StringUtils;
import org.apache.flink.api.common.eventtime.WatermarkStrategy;
import org.apache.flink.api.common.restartstrategy.RestartStrategies;
import org.apache.flink.configuration.Configuration;
import org.apache.flink.connector.kafka.source.KafkaSource;
import org.apache.flink.contrib.streaming.state.EmbeddedRocksDBStateBackend;
import org.apache.flink.streaming.api.CheckpointingMode;
import org.apache.flink.streaming.api.datastream.DataStream;
import org.apache.flink.streaming.api.datastream.DataStreamSource;
import org.apache.flink.streaming.api.datastream.SingleOutputStreamOperator;
import org.apache.flink.streaming.api.environment.CheckpointConfig;
import org.apache.flink.streaming.api.environment.StreamExecutionEnvironment;
import org.apache.flink.streaming.api.functions.ProcessFunction;
import org.apache.flink.streaming.api.windowing.time.Time;
import org.apache.flink.util.Collector;
import org.apache.flink.util.OutputTag;

import java.time.LocalDate;
import java.util.Properties;

@Slf4j
public class AdDayStreamJob {

    public static void main(String[] args) throws Exception {
        boolean isTest = false;
        StreamExecutionEnvironment env = StreamExecutionEnvironment.getExecutionEnvironment();
        // 任务失败后的重启策略
        // env.setRestartStrategy(RestartStrategies.fixedDelayRestart(0, 0));
        env.setRestartStrategy(RestartStrategies.noRestart());

        // 加载配置文件到 flink的全局配置中
        Properties props = new Properties();
        props.load(AdHourStreamJob.class.getResourceAsStream(isTest ? "/application.test.properties" : "/application.properties"));
        Configuration configuration = new Configuration();
        props.stringPropertyNames().forEach(key -> {
            String value = props.getProperty(key);
            configuration.setString(key.trim(), StringUtils.isBlank(value) ? "" : value.trim());
        });
        env.getConfig().setGlobalJobParameters(configuration);

        // checkpoint配置
        env.enableCheckpointing(5 * 60 * 1000L, CheckpointingMode.EXACTLY_ONCE);
        // checkpoint执行超时时间，超时则 checkpoint失败
        env.getCheckpointConfig().setCheckpointTimeout(5 * 60 * 1000L);
        // checkpoint执行最小间隔时间
        env.getCheckpointConfig().setMinPauseBetweenCheckpoints(60 * 1000L);
        // 允许并行执行 checkpoint个数，当指定了 minPauseBetweenCheckpoints时，其值无效（就是 1）
        env.getCheckpointConfig().setMaxConcurrentCheckpoints(1);
        // 开启 checkpoints的外部持久化，但是在job失败的时候不会自动清理，需要自己手工清理state。
        // ExternalizedCheckpointCleanup用于指定当job canceled的时候外部的 checkpoint该如何清理
        // DELETE_ON_CANCELLATION: 在job canceled的时候会自动删除外部 state，但是如果是FAILED的状态则会保留
        // RETAIN_ON_CANCELLATION：在job canceled的时候保留外部 state
        env.getCheckpointConfig().enableExternalizedCheckpoints(CheckpointConfig.ExternalizedCheckpointCleanup.RETAIN_ON_CANCELLATION);
        // 大概是允许 checkpoint失败几次，默认 0
        env.getCheckpointConfig().setTolerableCheckpointFailureNumber(0);
        env.setStateBackend(new EmbeddedRocksDBStateBackend(true));
        if (StringUtils.isNotBlank(props.getProperty(ApplicationProperties.FLINK_CHECKPOINT_SAVEPOINT))) {
            env.getCheckpointConfig().setCheckpointStorage(props.getProperty(ApplicationProperties.FLINK_CHECKPOINT_SAVEPOINT));
        }

        KafkaSource<String> adStreamOfDaySource = KafkaComponent.buildKafkaSource(isTest, props, KafkaComponent.KafkaTopic.adDayTopic,
                KafkaComponent.KafkaTopic.KafkaGroupId.adDayConsumerGroup);
        DataStreamSource<String> adStreamOfDayIn = env.fromSource(adStreamOfDaySource, WatermarkStrategy.noWatermarks(),
                "adDaySource_kafka");

        // 广告日数据。往前回滚 10天
        final OutputTag<AdDataOfDayODS> adDayStreamRollDayTag = new OutputTag<AdDataOfDayODS>("adDayStreamRollDayTag") {
        };
        // 广告日数据。往前回滚 1年
        final OutputTag<AdDataOfDayODS> adDayStreamRollYearTag = new OutputTag<AdDataOfDayODS>("adDayStreamRollYearTag") {
        };
        SingleOutputStreamOperator<AdStatOfDayODSDTO> adDayODSStream = adStreamOfDayIn.filter(StringUtils::isNotBlank)
                .map(AdStatOfDayODSDTO::byJson);

        // 拆分流
        SingleOutputStreamOperator<AdStatOfDayODSDTO> adDayODSStreamSplit = adDayODSStream.process(new ProcessFunction<AdStatOfDayODSDTO, AdStatOfDayODSDTO>() {
            @Override
            public void processElement(AdStatOfDayODSDTO adDataDTO, ProcessFunction<AdStatOfDayODSDTO, AdStatOfDayODSDTO>.Context context, Collector<AdStatOfDayODSDTO> collector) throws Exception {
                LocalDate startDate = adDataDTO.getStartDate();
                LocalDate endDate = adDataDTO.getEndDate();
                AdDataOfDayODS adODS = adDataDTO.getAdDataOfDayODS();
                if (DateUtil.intervalOfDays(startDate, endDate) > 31L) {
                    // 拉取时间间隔超过 1个月，账号回滚 365天的数据
                    context.output(adDayStreamRollYearTag, adODS);
                } else {
                    // 每日往前回滚 10天的数据
                    context.output(adDayStreamRollDayTag, adODS);
                }
            }
        });

        // 回滚 30天的数据计算
        SingleOutputStreamOperator<AdStatOfDayDWD> adDayDWDMonthStream = adDayODSStreamSplit.getSideOutput(adDayStreamRollDayTag)
                .keyBy(AdDataOfDayODS::getAdId)
                .process(new AdDayDWDRollMonthProcess())
                .setParallelism(3);

        // 单个账号回滚一年
        SingleOutputStreamOperator<AdStatOfDayDWD> adDayDWDYearStream = adDayODSStreamSplit.getSideOutput(adDayStreamRollYearTag)
                .keyBy(AdDataOfDayODS::getAdId)
                .process(new AdDayDWDRollYearProcess());

        DataStream<AdStatOfDayDWD> adDayStream = adDayDWDMonthStream.union(adDayDWDYearStream);
        // 写入 maxCompute
        new KeyedBatchStream<>(adDayStream, AdStatOfDayDWD::getStatDay, 3000L, Time.minutes(1L))
                .toBatch()
                .setParallelism(6)
                .addSink(new OssBatchStreamSink<>(AdStatOfDayDWD.class, new OssBatchStreamSink.MonitoringGenerateOssObjectName("ad_stat_of_day_dwd")))
                .setParallelism(6)
                .name("sink_ad_year_dwd");
        // 写入 mysql
        new BatchStream<>(adDayStream, 2000L, Time.minutes(1L))
                .toBatch()
                .setParallelism(1)
                .addSink(new AdDayDWDToDBBatchSink())
                .setParallelism(1)
                .name("sink_ad_day_for_db");

        env.execute(isTest ? "ad_day_stream_job_test" : "ad_day_stream_job");
    }
}
