package flink.zanxiangnet.ad.monitoring.maxcompute.bean;

import flink.zanxiangnet.ad.monitoring.maxcompute.bean.annotation.MaxComputeColumn;
import lombok.AllArgsConstructor;
import lombok.Builder;
import lombok.Data;
import lombok.NoArgsConstructor;
import org.apache.commons.lang3.StringUtils;

import java.lang.reflect.Field;
import java.lang.reflect.Method;
import java.lang.reflect.Modifier;
import java.nio.charset.StandardCharsets;
import java.util.*;

public class BeanUtil {

    /**
     * 解析 javabean的字段
     *
     * @param clazz
     * @return
     */
    public static List<FieldInfo> parseBeanField(Class<?> clazz) {
        List<Field> fieldList = new ArrayList<>();
        Class<?> tempClazz = clazz;
        while (tempClazz != null) {
            Field[] fields = tempClazz.getDeclaredFields();
            if (fields.length != 0) {
                fieldList.addAll(Arrays.asList(fields));
            }
            tempClazz = tempClazz.getSuperclass();
        }
        if (fieldList.size() == 0) {
            throw new RuntimeException("Failed parsed class[" + clazz.getName() + "]. No Field!");
        }
        List<FieldInfo> fieldInfoList = new ArrayList<>(fieldList.size());
        for (Field field : fieldList) {
            if (Modifier.isFinal(field.getModifiers()) || Modifier.isStatic(field.getModifiers())) {
                continue;
            }
            MaxComputeColumn annotation = field.getAnnotation(MaxComputeColumn.class);
            FieldInfo fieldInfo = new FieldInfo();
            fieldInfo.setField(field);
            fieldInfo.setFieldName(field.getName());
            try {
                fieldInfo.setGetMethod(methodOfGet(clazz, field));
                fieldInfo.setSetMethod(methodOfSet(clazz, field));
            } catch (NoSuchMethodException e) {
                throw new RuntimeException("Failed parsed class[" + clazz.getName() + "]. The get or set method of [" + field.getName() + "] was not found!");
            }
            if (annotation != null) {
                if (annotation.ignore()) {
                    continue;
                }
                if (annotation.isPartitioned()) {
                    fieldInfo.setUsePartitioned(true);
                }
                if (StringUtils.isNotBlank(annotation.value())) {
                    fieldInfo.setColumnName(annotation.value());
                } else {
                    fieldInfo.setColumnName(humpToLine(fieldInfo.getFieldName()));
                }
            } else {
                fieldInfo.setColumnName(humpToLine(fieldInfo.getFieldName()));
            }
            fieldInfoList.add(fieldInfo);
        }
        return fieldInfoList;
    }

    /**
     * 下划线转驼峰
     *
     * @param str
     * @return
     */
    public static String lineToHump(String str) {
        byte[] bytes = str.getBytes(StandardCharsets.UTF_8);
        byte[] result = new byte[str.length()];
        int index = 0;
        for (int i = 0; i < bytes.length; i++) {
            if (bytes[i] != 95) {
                result[index++] = bytes[i];
                continue;
            }
            if (i == bytes.length - 1) {
                result[index++] = bytes[i];
                continue;
            }
            byte nextChar = bytes[++i];
            if (97 <= nextChar && nextChar <= 122) {
                result[index++] = (byte) (nextChar - 32);
            } else {
                result[index++] = nextChar;
            }
        }
        return new String(result, 0, index);
    }

    /**
     * 驼峰转下划线
     *
     * @param str
     * @return
     */
    public static String humpToLine(String str) {
        byte[] bytes = str.getBytes(StandardCharsets.UTF_8);
        byte[] result = new byte[str.length() * 2];
        int index = 0;
        for (byte c : bytes) {
            if (65 <= c && c <= 90) {
                result[index++] = 95;
                result[index++] = (byte) (c + 32);
            } else {
                result[index++] = c;
            }
        }
        return new String(result, 0, index);
    }

    public static <T> Method methodOfGet(Class<T> clazz, Field field) throws NoSuchMethodException {
        String fieldName = field.getName();
        String methodName = "get" + fieldName.substring(0, 1).toUpperCase() + fieldName.substring(1);
        return clazz.getMethod(methodName);
    }

    public static <T> Method methodOfSet(Class<T> clazz, Field field) throws NoSuchMethodException {
        String fieldName = field.getName();
        String methodName = "set" + fieldName.substring(0, 1).toUpperCase() + fieldName.substring(1);
        return clazz.getMethod(methodName, field.getType());
    }

    @Data
    @NoArgsConstructor
    @AllArgsConstructor
    @Builder
    public static class FieldInfo {
        /**
         * 是否用于分
         */
        private boolean usePartitioned = false;
        /**
         * 字段名
         */
        private String fieldName;
        /**
         * 数据库中的列名
         */
        private String columnName;
        /**
         * 字段
         */
        private Field field;
        /**
         * 字段的 get方法
         */
        private Method getMethod;

        /**
         * 字段的 set方法
         */
        private Method setMethod;
    }
}
