package flink.zanxiangnet.ad.monitoring.process;

import flink.zanxiangnet.ad.monitoring.config.MysqlDataSourceFactory;
import flink.zanxiangnet.ad.monitoring.dao.mapper.AdStatOfDayDWDMapper;
import flink.zanxiangnet.ad.monitoring.pojo.entity.AdDataOfDayODS;
import flink.zanxiangnet.ad.monitoring.pojo.entity.AdStatOfDayDWD;
import flink.zanxiangnet.ad.monitoring.pojo.properties.ApplicationProperties;
import com.zanxiangnet.module.util.DateUtil;
import flink.zanxiangnet.ad.monitoring.util.StateTtlUtil;
import lombok.extern.slf4j.Slf4j;
import org.apache.flink.api.common.state.MapState;
import org.apache.flink.api.common.state.MapStateDescriptor;
import org.apache.flink.api.common.state.ValueState;
import org.apache.flink.api.common.state.ValueStateDescriptor;
import org.apache.flink.api.common.time.Time;
import org.apache.flink.configuration.Configuration;
import org.apache.flink.streaming.api.functions.KeyedProcessFunction;
import org.apache.flink.util.Collector;
import org.apache.ibatis.datasource.DataSourceFactory;
import org.apache.ibatis.mapping.Environment;
import org.apache.ibatis.session.SqlSession;
import org.apache.ibatis.session.SqlSessionFactory;
import org.apache.ibatis.session.SqlSessionFactoryBuilder;
import org.apache.ibatis.transaction.jdbc.JdbcTransactionFactory;

import java.time.LocalDate;
import java.util.Map;
import java.util.Properties;
import java.util.stream.Collectors;

@Slf4j
public class AdDayDWDRollMonthProcess extends KeyedProcessFunction<Long, AdDataOfDayODS, AdStatOfDayDWD> {
    private SqlSessionFactory sqlSessionFactory;
    // 上次查询的时间
    private ValueState<String> lastQueryDayState;
    // 之前聚合的昨天的数据
    private MapState<String, AdStatOfDayDWD> historyReduceState;

    @Override
    public void open(Configuration conf) {
        Map<String, String> params = getRuntimeContext()
                .getExecutionConfig()
                .getGlobalJobParameters()
                .toMap();

        Properties mysqlProps = new Properties();
        mysqlProps.setProperty(ApplicationProperties.MYSQL_DRIVER_CLASS_NAME, params.get(ApplicationProperties.MYSQL_DRIVER_CLASS_NAME));
        mysqlProps.setProperty(ApplicationProperties.MYSQL_URL, params.get(ApplicationProperties.MYSQL_URL));
        mysqlProps.setProperty(ApplicationProperties.MYSQL_USERNAME, params.get(ApplicationProperties.MYSQL_USERNAME));
        mysqlProps.setProperty(ApplicationProperties.MYSQL_PASSWORD, params.get(ApplicationProperties.MYSQL_PASSWORD));
        mysqlProps.setProperty(ApplicationProperties.MYSQL_INIT_POOL_SIZE, params.get(ApplicationProperties.MYSQL_INIT_POOL_SIZE));
        mysqlProps.setProperty(ApplicationProperties.MYSQL_MAX_POOL_SIZE, params.get(ApplicationProperties.MYSQL_MAX_POOL_SIZE));

        DataSourceFactory dataSourceFactory = new MysqlDataSourceFactory();
        dataSourceFactory.setProperties(mysqlProps);
        Environment environment = new Environment("mysql", new JdbcTransactionFactory(), dataSourceFactory.getDataSource());
        org.apache.ibatis.session.Configuration configuration = new org.apache.ibatis.session.Configuration(environment);
        // 开启驼峰规则
        configuration.setMapUnderscoreToCamelCase(true);
        configuration.getTypeAliasRegistry().registerAlias(AdStatOfDayDWD.class);
        // addMapper一定要放到 alias的后面！！！！！
        configuration.addMapper(AdStatOfDayDWDMapper.class);
        sqlSessionFactory = new SqlSessionFactoryBuilder().build(configuration);

        ValueStateDescriptor<String> lastQueryDayStateDescriptor = new ValueStateDescriptor<>("lastQueryDayState", String.class);
        lastQueryDayStateDescriptor.enableTimeToLive(StateTtlUtil.rocketDBTtl(Time.hours(25)));
        lastQueryDayState = getRuntimeContext().getState(lastQueryDayStateDescriptor);

        MapStateDescriptor<String, AdStatOfDayDWD> historyReduceStateDescriptor = new MapStateDescriptor<>("historyReduceState", String.class, AdStatOfDayDWD.class);
        historyReduceStateDescriptor.enableTimeToLive(StateTtlUtil.rocketDBTtl(Time.hours(25)));
        historyReduceState = getRuntimeContext().getMapState(historyReduceStateDescriptor);
    }

    @Override
    public void processElement(AdDataOfDayODS element, KeyedProcessFunction<Long, AdDataOfDayODS, AdStatOfDayDWD>.Context context, Collector<AdStatOfDayDWD> collector) throws Exception {
        LocalDate statDay = DateUtil.parseLocalDate(element.getStatDay());
        long now = System.currentTimeMillis();
        long adId = element.getAdId();

        String lastQueryDay = lastQueryDayState.value();
        // 从 maxCompute查找广告的历史数据
        if (lastQueryDay == null || !lastQueryDay.equals(DateUtil.formatLocalDate(LocalDate.now()))) {
            try (SqlSession session = sqlSessionFactory.openSession()) {
                AdStatOfDayDWDMapper mapper = session.getMapper(AdStatOfDayDWDMapper.class);
                Map<String, AdStatOfDayDWD> historyData = mapper.lastReduceResult(adId, DateUtil.formatLocalDate(LocalDate.now().minusDays(60L)), DateUtil.formatLocalDate(LocalDate.now().minusDays(1L)), 60).stream()
                        .collect(Collectors.toMap(AdStatOfDayDWD::getStatDay, data -> data, (val1, val2) -> val2));
                historyReduceState.clear();
                historyReduceState.putAll(historyData);
                lastQueryDayState.update(DateUtil.formatLocalDate(LocalDate.now()));
            }
        }
        AdStatOfDayDWD lastReduceData = null;
        for (int i = 1; i <= 60; i++) {
            lastReduceData = historyReduceState.get(DateUtil.formatLocalDate(statDay.minusDays(i)));
            if (lastReduceData != null) {
                break;
            }
        }
        AdStatOfDayDWD newStatData = AdStatOfDayDWD.reduce(lastReduceData, element, now);
        historyReduceState.put(DateUtil.formatLocalDate(statDay), newStatData);
        collector.collect(newStatData);
    }
}
