package flink.zanxiangnet.ad.monitoring.process;

import flink.zanxiangnet.ad.monitoring.config.MysqlDataSourceFactory;
import flink.zanxiangnet.ad.monitoring.dao.mapper.AdStatOfDayDWDMapper;
import flink.zanxiangnet.ad.monitoring.pojo.entity.AdStatOfDayDWD;
import flink.zanxiangnet.ad.monitoring.pojo.properties.ApplicationProperties;
import com.zanxiangnet.module.util.DateUtil;
import lombok.extern.slf4j.Slf4j;
import org.apache.flink.api.common.state.ValueState;
import org.apache.flink.api.common.state.ValueStateDescriptor;
import org.apache.flink.configuration.Configuration;
import org.apache.flink.streaming.api.functions.KeyedProcessFunction;
import org.apache.flink.util.Collector;
import org.apache.ibatis.datasource.DataSourceFactory;
import org.apache.ibatis.mapping.Environment;
import org.apache.ibatis.session.SqlSession;
import org.apache.ibatis.session.SqlSessionFactory;
import org.apache.ibatis.session.SqlSessionFactoryBuilder;
import org.apache.ibatis.transaction.jdbc.JdbcTransactionFactory;

import java.time.LocalDate;
import java.util.List;
import java.util.Map;
import java.util.Properties;

/**
 * 小时流数据补全，把中间没消耗的时间段填充 0
 */
@Slf4j
public class AdDayOnTimeStreamCompletionProcess extends KeyedProcessFunction<Long, AdStatOfDayDWD, AdStatOfDayDWD> {

    private SqlSessionFactory sqlSessionFactory;
    // 上次查询的时间
    private ValueState<String> lastQueryDayState;

    @Override
    public void open(Configuration conf) {
        Map<String, String> params = getRuntimeContext()
                .getExecutionConfig()
                .getGlobalJobParameters()
                .toMap();

        Properties mysqlProps = new Properties();
        mysqlProps.setProperty(ApplicationProperties.MYSQL_DRIVER_CLASS_NAME, params.get(ApplicationProperties.MYSQL_DRIVER_CLASS_NAME));
        mysqlProps.setProperty(ApplicationProperties.MYSQL_URL, params.get(ApplicationProperties.MYSQL_URL));
        mysqlProps.setProperty(ApplicationProperties.MYSQL_USERNAME, params.get(ApplicationProperties.MYSQL_USERNAME));
        mysqlProps.setProperty(ApplicationProperties.MYSQL_PASSWORD, params.get(ApplicationProperties.MYSQL_PASSWORD));
        mysqlProps.setProperty(ApplicationProperties.MYSQL_INIT_POOL_SIZE, params.get(ApplicationProperties.MYSQL_INIT_POOL_SIZE));
        mysqlProps.setProperty(ApplicationProperties.MYSQL_MAX_POOL_SIZE, params.get(ApplicationProperties.MYSQL_MAX_POOL_SIZE));

        DataSourceFactory dataSourceFactory = new MysqlDataSourceFactory();
        dataSourceFactory.setProperties(mysqlProps);
        Environment environment = new Environment("mysql", new JdbcTransactionFactory(), dataSourceFactory.getDataSource());
        org.apache.ibatis.session.Configuration configuration = new org.apache.ibatis.session.Configuration(environment);
        // 开启驼峰规则
        configuration.setMapUnderscoreToCamelCase(true);
        configuration.getTypeAliasRegistry().registerAlias(AdStatOfDayDWD.class);
        // addMapper一定要放到 alias的后面！！！！！
        configuration.addMapper(AdStatOfDayDWDMapper.class);
        sqlSessionFactory = new SqlSessionFactoryBuilder().build(configuration);

        ValueStateDescriptor<String> lastQueryDayStateDescriptor = new ValueStateDescriptor<>("lastQueryDayState", String.class);
        lastQueryDayState = getRuntimeContext().getState(lastQueryDayStateDescriptor);
    }

    @Override
    public void processElement(AdStatOfDayDWD adStatOfDayDWD, KeyedProcessFunction<Long, AdStatOfDayDWD, AdStatOfDayDWD>.Context context, Collector<AdStatOfDayDWD> collector) throws Exception {
        String lastQueryTime = lastQueryDayState.value();
        if (lastQueryTime != null && lastQueryTime.compareTo(adStatOfDayDWD.getStatDay()) >= 0) {
            collector.collect(adStatOfDayDWD);
            return;
        }
        lastQueryDayState.update(adStatOfDayDWD.getStatDay());
        LocalDate statDay = DateUtil.parseLocalDate(adStatOfDayDWD.getStatDay());
        try (SqlSession session = sqlSessionFactory.openSession()) {
            AdStatOfDayDWDMapper mapper = session.getMapper(AdStatOfDayDWDMapper.class);
            List<AdStatOfDayDWD> lastReduceList = mapper.lastReduceResult(adStatOfDayDWD.getAdId(), null, DateUtil.formatLocalDate(statDay.minusDays(1L)), 1);
            if (lastReduceList.isEmpty()) {
                collector.collect(adStatOfDayDWD);
                return;
            }
            AdStatOfDayDWD lastReduce = lastReduceList.get(0);
            LocalDate statDate = DateUtil.parseLocalDate(adStatOfDayDWD.getStatDay());
            LocalDate lastStatDate = DateUtil.parseLocalDate(lastReduce.getStatDay());
            if (lastStatDate.compareTo(statDate) >= 0) {
                collector.collect(adStatOfDayDWD);
                return;
            }
            long days = DateUtil.intervalOfDays(lastStatDate, statDate);
            if (days > 1) {
                // 中间有没数据的时间段，需要进行数据填充
                for (int i = 1; i < days; i++) {
                    // 要填充的时间
                    LocalDate completionDay = lastStatDate.plusDays(i);
                    collector.collect(AdStatOfDayDWD.completion(completionDay, lastReduce));
                }
            }
            collector.collect(adStatOfDayDWD);
        }
    }
}
