package flink.zanxiangnet.ad.monitoring.process;

import flink.zanxiangnet.ad.monitoring.config.MysqlDataSourceFactory;
import flink.zanxiangnet.ad.monitoring.dao.mapper.AdStatOfDayDWDMapper;
import flink.zanxiangnet.ad.monitoring.pojo.entity.AdDataOfHourODS;
import flink.zanxiangnet.ad.monitoring.pojo.entity.AdStatOfDayDWD;
import flink.zanxiangnet.ad.monitoring.pojo.entity.AdStatOfHourDWD;
import flink.zanxiangnet.ad.monitoring.pojo.properties.ApplicationProperties;
import com.zanxiangnet.module.util.DateUtil;
import flink.zanxiangnet.ad.monitoring.util.StateTtlUtil;
import lombok.extern.slf4j.Slf4j;
import org.apache.flink.api.common.state.*;
import org.apache.flink.api.common.time.Time;
import org.apache.flink.api.common.typeinfo.Types;
import org.apache.flink.configuration.Configuration;
import org.apache.flink.streaming.api.functions.KeyedProcessFunction;
import org.apache.flink.util.Collector;
import org.apache.ibatis.datasource.DataSourceFactory;
import org.apache.ibatis.mapping.Environment;
import org.apache.ibatis.session.SqlSession;
import org.apache.ibatis.session.SqlSessionFactory;
import org.apache.ibatis.session.SqlSessionFactoryBuilder;
import org.apache.ibatis.transaction.jdbc.JdbcTransactionFactory;

import java.time.LocalDate;
import java.util.*;

/**
 * 往前回滚 10天的小时数据（警告，此数据依赖小时数据必定在天数据之后拉取来实现的。不然会拉下昨天的数据没有统计进来）
 */
@Slf4j
public class AdHourDWDProcess extends KeyedProcessFunction<Long, AdDataOfHourODS, AdStatOfHourDWD> {
    private SqlSessionFactory sqlSessionFactory;
    // 上次查询的天数据
    private ValueState<Long> lastQueryTimeState;
    // 聚合的天的数据
    private ListState<AdStatOfDayDWD> historyReduceState;
    // 上小时聚合的结果
    private MapState<String, Map<Integer, AdDataOfHourODS>> historyState;

    @Override
    public void open(Configuration conf) {
        Map<String, String> params = getRuntimeContext()
                .getExecutionConfig()
                .getGlobalJobParameters()
                .toMap();

        Properties mysqlProps = new Properties();
        mysqlProps.setProperty(ApplicationProperties.MYSQL_DRIVER_CLASS_NAME, params.get(ApplicationProperties.MYSQL_DRIVER_CLASS_NAME));
        mysqlProps.setProperty(ApplicationProperties.MYSQL_URL, params.get(ApplicationProperties.MYSQL_URL));
        mysqlProps.setProperty(ApplicationProperties.MYSQL_USERNAME, params.get(ApplicationProperties.MYSQL_USERNAME));
        mysqlProps.setProperty(ApplicationProperties.MYSQL_PASSWORD, params.get(ApplicationProperties.MYSQL_PASSWORD));
        mysqlProps.setProperty(ApplicationProperties.MYSQL_INIT_POOL_SIZE, params.get(ApplicationProperties.MYSQL_INIT_POOL_SIZE));
        mysqlProps.setProperty(ApplicationProperties.MYSQL_MAX_POOL_SIZE, params.get(ApplicationProperties.MYSQL_MAX_POOL_SIZE));

        DataSourceFactory dataSourceFactory = new MysqlDataSourceFactory();
        dataSourceFactory.setProperties(mysqlProps);
        Environment environment = new Environment("mysql", new JdbcTransactionFactory(), dataSourceFactory.getDataSource());
        org.apache.ibatis.session.Configuration configuration = new org.apache.ibatis.session.Configuration(environment);
        // 开启驼峰规则
        configuration.setMapUnderscoreToCamelCase(true);
        configuration.getTypeAliasRegistry().registerAlias(AdStatOfDayDWD.class);
        // addMapper一定要放到 alias的后面！！！！！
        configuration.addMapper(AdStatOfDayDWDMapper.class);
        sqlSessionFactory = new SqlSessionFactoryBuilder().build(configuration);

        ValueStateDescriptor<Long> lastQueryTimeStateDescriptor = new ValueStateDescriptor<>("lastQueryTimeState", Types.LONG);
        lastQueryTimeStateDescriptor.enableTimeToLive(StateTtlUtil.rocketDBTtl(Time.hours(1)));
        lastQueryTimeState = getRuntimeContext().getState(lastQueryTimeStateDescriptor);
        ListStateDescriptor<AdStatOfDayDWD> historyReduceStateDescriptor = new ListStateDescriptor<>("historyReduceState", Types.POJO(AdStatOfDayDWD.class));
        historyReduceStateDescriptor.enableTimeToLive(StateTtlUtil.rocketDBTtl(Time.hours(1)));
        historyReduceState = getRuntimeContext().getListState(historyReduceStateDescriptor);
        MapStateDescriptor<String, Map<Integer, AdDataOfHourODS>> historyStateDescriptor = new MapStateDescriptor<>("historyState", Types.STRING, Types.MAP(Types.INT, Types.POJO(AdDataOfHourODS.class)));
        historyStateDescriptor.enableTimeToLive(StateTtlUtil.rocketDBTtl(Time.hours(24)));
        historyState = getRuntimeContext().getMapState(historyStateDescriptor);
    }

    @Override
    public void processElement(AdDataOfHourODS element, KeyedProcessFunction<Long, AdDataOfHourODS, AdStatOfHourDWD>.Context context, Collector<AdStatOfHourDWD> collector) throws Exception {
        LocalDate statDay = DateUtil.parseLocalDate(element.getStatDay());
        long now = System.currentTimeMillis();
        long adId = element.getAdId();

        // 每小时从 ck更新下历史的天数据
        Long lastQueryTime = lastQueryTimeState.value();
        if (lastQueryTime == null || (now - lastQueryTime > 60 * 60 * 1000)) {
            try (SqlSession session = sqlSessionFactory.openSession()) {
                AdStatOfDayDWDMapper mapper = session.getMapper(AdStatOfDayDWDMapper.class);
                List<AdStatOfDayDWD> historyData = mapper.lastReduceResult(adId, null, DateUtil.formatLocalDate(LocalDate.now().minusDays(1L)), 60);
                historyReduceState.clear();
                historyReduceState.update(historyData);
                lastQueryTimeState.update(now);
            }
        }
        AdStatOfDayDWD yesterdayReduceData = null;
        for (AdStatOfDayDWD dayDWD : historyReduceState.get()) {
            if (dayDWD.getStatDay().compareTo(element.getStatDay()) < 0) {
                yesterdayReduceData = dayDWD;
                break;
            }
        }

        Map<Integer, AdDataOfHourODS> dayODSHourMap = historyState.get(element.getStatDay());
        List<AdDataOfHourODS> dayODSList = new ArrayList<>();
        dayODSList.add(element);
        if (dayODSHourMap == null) {
            dayODSHourMap = new HashMap<>(24);
        }
        if (!dayODSHourMap.isEmpty()) {
            dayODSHourMap.values().forEach(hourODS -> {
                if (hourODS.getHour() < element.getHour()) {
                    dayODSList.add(hourODS);
                }
            });
        }

        AdStatOfHourDWD newStatData = AdStatOfHourDWD.reduce(yesterdayReduceData, dayODSList, element, now);
        collector.collect(newStatData);

        dayODSHourMap.put(element.getHour(), element);
        historyState.put(element.getStatDay(), dayODSHourMap);

        // 往前清理 15天的数据
        /*historyState.remove(DateUtil.formatLocalDate(statDay.minusDays(15L)));
        historyState.remove(DateUtil.formatLocalDate(statDay.minusDays(16L)));
        historyState.remove(DateUtil.formatLocalDate(statDay.minusDays(17L)));*/
    }
}
