package flink.zanxiangnet.ad.monitoring.sink;

import com.aliyun.odps.Odps;
import com.aliyun.odps.PartitionSpec;
import com.aliyun.odps.account.Account;
import com.aliyun.odps.account.AliyunAccount;
import com.aliyun.odps.data.Record;
import com.aliyun.odps.tunnel.TableTunnel;
import com.aliyun.odps.tunnel.TunnelException;
import com.aliyun.oss.OSS;
import com.aliyun.oss.OSSClientBuilder;
import com.aliyun.oss.model.PutObjectResult;
import com.zanxiangnet.module.util.DateUtil;
import com.zanxiangnet.module.util.JsonUtil;
import com.zanxiangnet.module.util.csv.CsvUtil;
import flink.zanxiangnet.ad.monitoring.maxcompute.MaxComputeLog;
import flink.zanxiangnet.ad.monitoring.maxcompute.bean.BeanUtil;
import flink.zanxiangnet.ad.monitoring.maxcompute.bean.annotation.MaxComputeTable;
import flink.zanxiangnet.ad.monitoring.pojo.properties.ApplicationProperties;
import lombok.extern.slf4j.Slf4j;
import org.apache.commons.lang3.StringUtils;
import org.apache.flink.configuration.Configuration;
import org.apache.flink.streaming.api.functions.sink.RichSinkFunction;
import org.springframework.util.CollectionUtils;

import java.io.*;
import java.lang.reflect.InvocationTargetException;
import java.lang.reflect.Method;
import java.math.BigDecimal;
import java.time.LocalDate;
import java.time.LocalDateTime;
import java.time.format.DateTimeFormatter;
import java.util.List;
import java.util.Map;
import java.util.UUID;
import java.util.stream.Collectors;

/**
 * 批量数据写出到 OSS的 CSV文件
 *
 * @param <IN>
 */
@Slf4j
public class OssBatchStreamSink<T, IN extends List<T>> extends RichSinkFunction<IN> {
    private static final int RETRY_COUNT = 3;
    private static final DateTimeFormatter YYYY_MM_DD = DateTimeFormatter.ofPattern("yyyy_MM_dd");

    private final Class<T> clazz;

    private volatile transient OSS ossClient;
    private String endpoint;
    // 暂存的 oss地址
    private String ossBucket;

    private GenerateOssObjectName generateOssObjectName;

    public OssBatchStreamSink(Class<T> clazz, GenerateOssObjectName generateOssObjectName) {
        this.clazz = clazz;
        this.generateOssObjectName = generateOssObjectName;
    }

    @Override
    public void open(Configuration config) {
        Map<String, String> params = getRuntimeContext()
                .getExecutionConfig()
                .getGlobalJobParameters()
                .toMap();
        endpoint = params.get(ApplicationProperties.OSS_ENDPOINT);
        ossClient = new OSSClientBuilder().build(params.get(ApplicationProperties.OSS_ENDPOINT),
                params.get(ApplicationProperties.OSS_ASSESS_KEY_ID),
                params.get(ApplicationProperties.OSS_ASSESS_KEY_SECRET));
        ossBucket = params.get(ApplicationProperties.OSS_BUCKET);
    }

    /**
     * 将值写入到 Sink。每个值都会调用此函数
     *
     * @param value
     * @param context
     */
    @Override
    public void invoke(IN value, Context context) {
        LocalDateTime now = LocalDateTime.now();
        long start = System.currentTimeMillis();
        int index = 0;
        do {
            String savePath = System.getProperty("java.io.tmpdir") + "/flink/oss/" + now.toLocalDate().format(YYYY_MM_DD) + "/";
            File path = new File(savePath);
            if (!path.exists()) {
                path.mkdirs();
            }
            String fileName = UUID.randomUUID() + ".csv";
            File file = new File(savePath + fileName);
            try {
                CsvUtil.print(clazz, value, new FileWriter(file));
                String objectName = generateOssObjectName.objectName(now, savePath, fileName);
                PutObjectResult result = ossClient.putObject(ossBucket, objectName, new FileInputStream(file));
                System.out.println("OSS文件上传成功：" + "https://" + ossBucket +"." + endpoint + "/" + objectName);
                break;
            } catch (InstantiationException | IllegalAccessException | IOException e) {
                log.error("第 " + index + "次上传数据到 oss失败！" + e.getMessage(), e);
            } finally {
                file.delete();
            }
        } while (index++ < RETRY_COUNT);
        long end = System.currentTimeMillis();
        System.out.println("写入 " + value.size() + "条数据用时：" + (end - start));
    }

    @Override
    public void close() throws Exception {
        super.close();
    }

    public interface GenerateOssObjectName {

        String objectName(LocalDateTime executeTime, String savePath, String fileName);
    }

    public static class MonitoringGenerateOssObjectName implements GenerateOssObjectName, Serializable {
        private static final long serialVersionUID = 1L;

        private final String key;

        public MonitoringGenerateOssObjectName(String key) {
            this.key = key;
        }

        @Override
        public String objectName(LocalDateTime executeTime, String savePath, String fileName) {
            return "flink/oss/" + key + "/" + DateUtil.formatLocalDate(executeTime.toLocalDate()) + "/" + fileName;
        }
    }
}
