package flink.zanxiangnet.ad.monitoring.trigger;

import lombok.extern.slf4j.Slf4j;
import org.apache.flink.api.common.state.ReducingState;
import org.apache.flink.api.common.state.ReducingStateDescriptor;
import org.apache.flink.api.common.typeutils.base.LongSerializer;
import org.apache.flink.streaming.api.windowing.time.Time;
import org.apache.flink.streaming.api.windowing.triggers.Trigger;
import org.apache.flink.streaming.api.windowing.triggers.TriggerResult;
import org.apache.flink.streaming.api.windowing.windows.TimeWindow;

/**
 * 定时 && 定量触发
 *
 * @param <T>
 * @param <W>
 */
@Slf4j
public class TimerCountTrigger<T, W extends TimeWindow> extends Trigger<T, W> {

    private final ReducingStateDescriptor<Long> countStateDesc =
            new ReducingStateDescriptor<>("count", (Long::sum), LongSerializer.INSTANCE);

    // 缓存的最大数据量
    private final Long maxBufferCount;

    private final long bufferRefreshTime;

    public TimerCountTrigger(Long maxBufferCount, Time bufferRefreshTime) {
        this.maxBufferCount = maxBufferCount;
        this.bufferRefreshTime = bufferRefreshTime.toMilliseconds();
    }

    /**
     * 每次元素进入时调用
     *
     * @param element 元素
     * @param time    元素进入的时间（eventTime）
     * @param window  元素被添加到的窗口对象
     * @param tCtx    trigger上下文
     * @return
     * @throws Exception
     */
    @Override
    public TriggerResult onElement(T element, long time, W window, TriggerContext tCtx) throws Exception {
        ReducingState<Long> windowCountState = tCtx.getPartitionedState(countStateDesc);
        windowCountState.add(1L);
        if (windowCountState.get() >= maxBufferCount) {
            clear(window, tCtx);
            return TriggerResult.FIRE_AND_PURGE;
        }

        // 注册定时清理器
        tCtx.registerProcessingTimeTimer(window.getStart() + bufferRefreshTime);
        if (time >= window.getEnd()) {
            clear(window, tCtx);
            return TriggerResult.FIRE_AND_PURGE;
        }
        return TriggerResult.CONTINUE;
    }

    /**
     * trigger上下文设置的 processTimer触发时调用
     *
     * @param time   计时器触发的时间戳
     * @param window
     * @param tCtx
     * @return
     * @throws Exception
     */
    @Override
    public TriggerResult onProcessingTime(long time, W window, TriggerContext tCtx) throws Exception {
        clear(window, tCtx);
        return TriggerResult.FIRE_AND_PURGE;
    }

    /**
     * trigger上下文设置的 eventTimer触发时调用
     *
     * @param time
     * @param window
     * @param tCtx
     * @return
     * @throws Exception
     */
    @Override
    public TriggerResult onEventTime(long time, W window, TriggerContext tCtx) throws Exception {
        clear(window, tCtx);
        return TriggerResult.FIRE_AND_PURGE;
    }

    /**
     * 窗口被清除时调用
     *
     * @param w
     * @param tCtx
     * @throws Exception
     */
    @Override
    public void clear(W w, TriggerContext tCtx) throws Exception {
        tCtx.getPartitionedState(countStateDesc).clear();
        tCtx.deleteProcessingTimeTimer(w.getStart() + bufferRefreshTime);
    }

    @Override
    public boolean canMerge() {
        return true;
    }

    @Override
    public void onMerge(W window, OnMergeContext ctx) throws Exception {
        ctx.mergePartitionedState(countStateDesc);
    }
}
